import requests
from typing import Union
import aiohttp

from .ratelimit import RateLimiter


class HttpClient:
    def __init__(self, calls: int, period: int, base_url: str = "https://api.dexscreener.io/latest"):
        self._limiter = RateLimiter(calls, period)
        self.base_url = base_url
        
    def _create_absolute_url(self, relative: str) -> str:
        return f"{self.base_url}/{relative}"

    def request(self, method, url, **kwargs) -> Union[list, dict]:
        url = self._create_absolute_url(url)

        with self._limiter:
            r = requests.request(method, url, **kwargs)

            return r.json()

    async def request_async(self, method, url, **kwargs):
        url = self._create_absolute_url(url)

        async with self._limiter:
            async with aiohttp.ClientSession() as session:
                async with session.request(method, url, **kwargs) as response:
                    return await response.json()




import time
import threading
import collections
import asyncio


class RateLimiter:
    def __init__(self, max_calls, period):
        self.calls = collections.deque()

        self.period = period
        self.max_calls = max_calls

        self.sync_lock = threading.Lock()
        self.async_lock = asyncio.Lock()

    def __enter__(self):
        with self.sync_lock:
            sleep_time = self.get_sleep_time()

            if sleep_time > 0:
                time.sleep(sleep_time)

            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.sync_lock:
            self._clear_calls()

    async def __aenter__(self):
        async with self.async_lock:
            sleep_time = self.get_sleep_time()

            if sleep_time > 0:
                await asyncio.sleep(sleep_time)

            return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        async with self.async_lock:
            self._clear_calls()

    def get_sleep_time(self) -> float:
        if len(self.calls) >= self.max_calls:
            until = time.time() + self.period - self._timespan
            return until - time.time()

        return 0

    def _clear_calls(self):
        self.calls.append(time.time())

        while self._timespan >= self.period:
            self.calls.popleft()

    @property
    def _timespan(self):
        return self.calls[-1] - self.calls[0]


from .models import TokenPair, TokenInfo, OrderInfo
from .http_client import HttpClient
from typing import Optional, Iterable, List
import json


class DexscreenerClient:
    def __init__(self) -> None:
        self._client_60rpm: HttpClient = HttpClient(60, 60, base_url="https://api.dexscreener.com")
        self._client_300rpm: HttpClient = HttpClient(300, 60, base_url="https://api.dexscreener.com/latest")

    def get_latest_token_profiles(self) -> list[TokenInfo]:
        """
        Get the latest token profiles

        https://api.dexscreener.com/token-profiles/latest/v1

        :return:
            Response as list of TokenInfo model
        """
        resp = self._client_60rpm.request("GET", "token-profiles/latest/v1")
        return [TokenInfo(**token) for token in resp]

    async def get_latest_token_profiles_async(self) -> list[TokenInfo]:
        """
        Async version of `get_latest_token_profiles`
        """
        resp = await self._client_60rpm.request_async("GET", "token-profiles/latest/v1")
        return [TokenInfo(**token) for token in resp]

    def get_latest_boosted_tokens(self) -> list[TokenInfo]:
        """
        Get the latest boosted tokens

        https://api.dexscreener.com/token-boosts/latest/v1

        :return:
            Response as list of TokenInfo model
        """
        resp = self._client_60rpm.request("GET", "token-boosts/latest/v1")
        return [TokenInfo(**token) for token in resp]
        
    async def get_latest_boosted_tokens_async(self) -> list[TokenInfo]:
        """
        Async version of `get_latest_boosted_tokens`
        """
        resp = await self._client_60rpm.request_async("GET", "token-boosts/latest/v1")
        return [TokenInfo(**token) for token in resp]

    def get_tokens_most_active(self) -> list[TokenInfo]:
        """
        Get the tokens with most active boosts

        https://api.dexscreener.com/token-boosts/top/v1

        :return:
            Response as list of TokenInfo model
        """
        resp = self._client_60rpm.request("GET", "token-boosts/top/v1")
        return [TokenInfo(**token) for token in resp]

    async def get_tokens_most_active_async(self) -> list[TokenInfo]:
        """
        Async version of `get_tokens_most_active`
        """
        resp = await self._client_60rpm.request_async("GET", "token-boosts/top/v1")
        return [TokenInfo(**token) for token in resp]

    def get_orders_paid_of_token(self, chain_id: str, token_address: str) -> list[OrderInfo]:
        """
        Check orders paid for of token

        https://api.dexscreener.com/orders/v1/solana/A55XjvzRU4KtR3Lrys8PpLZQvPojPqvnv5bJVHMYy3Jv

        :return:
            Response as list of OrderInfo model
        """
        resp = self._client_60rpm.request("GET", f"orders/v1/{chain_id}/{token_address}")
        return [OrderInfo(**order) for order in resp]
    
    async def get_orders_paid_of_token_async(self, chain_id: str, token_address: str) -> list[OrderInfo]:
        """
        Async version of `get_orders_paid_of_token`
        """
        resp = await self._client_60rpm.request_async("GET", f"orders/v1/{chain_id}/{token_address}")
        return [OrderInfo(**order) for order in resp]

    def get_token_pair(self, chain: str, address: str) -> Optional[TokenPair]:
        """
        Fetch a pair on the provided chain id

        https://api.dexscreener.com/latest/dex/pairs/bsc/0x7213a321F1855CF1779f42c0CD85d3D95291D34C

        :param chain: Chain id
        :param address: Token address
        :return:
            Response as TokenPair model
        """
        resp = self._client_300rpm.request("GET", f"dex/pairs/{chain}/{address}")
        return TokenPair(**resp["pair"]) if resp["pair"] else None

    async def get_token_pair_async(self, chain: str, address: str) -> Optional[TokenPair]:
        """
        Async version of `get_token_pair`
        """
        resp = await self._client_300rpm.request_async("GET", f"dex/pairs/{chain}/{address}")        
        return TokenPair(**resp["pair"]) if resp["pair"] else None

    def get_token_pair_list(self, chain: str, addresses: Iterable[str]) -> List[TokenPair]:
        """
        Fetch multiple pairs on the provided chain id

        https://api.dexscreener.com/latest/dex/pairs/ethereum/0xC2aDdA861F89bBB333c90c492cB837741916A225,0x7BeA39867e4169DBe237d55C8242a8f2fcDcc387
        
        :param chain: Chain id
        :param addresses: Iterable of token addresses (up to 30)
        :return:
            Response as list of TokenPair models
        """
        addresses_list = list(addresses)
        if len(addresses_list) > 30:
            raise ValueError("The maximum number of addresses allowed is 30.")
        resp = self._client_300rpm.request("GET", f"dex/pairs/{chain}/{','.join(addresses_list)}")
        return [TokenPair(**pair) for pair in resp.get("pairs", [])]

    async def get_token_pair_list_async(self, chain: str, addresses: Iterable[str]) -> List[TokenPair]:
        """
        Async version of `get_token_pairs`
        """
        addresses_list = list(addresses)
        if len(addresses_list) > 30:
            raise ValueError("The maximum number of addresses allowed is 30.")
        resp = await self._client_300rpm.request_async("GET", f"dex/pairs/{chain}/{','.join(addresses_list)}")
        return [TokenPair(**pair) for pair in resp.get("pairs", [])]    
    
    def get_token_pairs(self, address: str) -> list[TokenPair]:
        """
        Get pairs matching base token address

        https://api.dexscreener.com/latest/dex/tokens/0x2170Ed0880ac9A755fd29B2688956BD959F933F8

        :param address: Token address
        :return:
            Response as list of TokenPair model
        """
        resp = self._client_300rpm.request("GET",  f"dex/tokens/{address}")        
        return [TokenPair(**pair) for pair in resp.get("pairs", [])]

    async def get_token_pairs_async(self, address: str) -> list[TokenPair]:
        """
        Async version of `get_token_pairs`
        """
        resp = await self._client_300rpm.request_async("GET", f"dex/tokens/{address}")        
        return [TokenPair(**pair) for pair in resp.get("pairs", [])]

    def search_pairs(self, search_query: str) -> list[TokenPair]:
        """
        Search for pairs matching query

        https://api.dexscreener.com/latest/dex/tokens/0x2170Ed0880ac9A755fd29B2688956BD959F933F8

        :param search_query: query (e.g.: WBTC or WBTC/USDC)
        :return:
            Response as list of TokenPair model
        """
        resp = self._client_300rpm.request("GET", f"dex/search/?q={search_query}")        
        return [TokenPair(**pair) for pair in resp.get("pairs", [])]

    async def search_pairs_async(self, search_query: str) -> list[TokenPair]:
        """
        Async version of `search_pairs`
        """
        resp = await self._client_300rpm.request_async("GET", f"dex/search/?q={search_query}")        
        return [TokenPair(**pair) for pair in resp.get("pairs", [])]



from pydantic import BaseModel, Field
from typing import Optional
import datetime as dt


class BaseToken(BaseModel):
    address: str
    name: str
    symbol: str


class TransactionCount(BaseModel):
    buys: int
    sells: int


class PairTransactionCounts(BaseModel):
    m5: TransactionCount
    h1: TransactionCount
    h6: TransactionCount
    h24: TransactionCount


class _TimePeriodsFloat(BaseModel):
    m5: Optional[float] = 0.0
    h1: Optional[float] = 0.0
    h6: Optional[float] = 0.0
    h24: Optional[float] = 0.0


class VolumeChangePeriods(_TimePeriodsFloat):
    ...


class PriceChangePeriods(_TimePeriodsFloat):
    ...


class Liquidity(BaseModel):
    usd: Optional[float] = None
    base: float
    quote: float


class TokenPair(BaseModel):
    chain_id: str = Field(..., alias="chainId")
    dex_id: str = Field(..., alias="dexId")
    url: str = Field(...)
    pair_address: str = Field(..., alias="pairAddress")
    base_token: BaseToken = Field(..., alias="baseToken")
    quote_token: BaseToken = Field(..., alias="quoteToken")
    price_native: float = Field(..., alias="priceNative")
    price_usd: Optional[float] = Field(None, alias="priceUsd")
    transactions: PairTransactionCounts = Field(..., alias="txns")
    volume: VolumeChangePeriods
    price_change: PriceChangePeriods = Field(..., alias="priceChange")
    liquidity: Optional[Liquidity] = None
    fdv: Optional[float] = 0.0
    pair_created_at: Optional[dt.datetime] = Field(None, alias="pairCreatedAt")


class TokenLink(BaseModel):
    type: Optional[str] = None
    label: Optional[str] = None
    url: Optional[str] = None


class TokenInfo(BaseModel):
    url: str
    chain_id: str = Field(..., alias="chainId")
    token_address: str = Field(..., alias="tokenAddress")
    amount: float = 0.0 # Not sure if this is the best logic
    total_amount: float = Field(0.0, alias="totalAmount")
    icon: Optional[str] = None
    header: Optional[str] = None
    description: Optional[str] = None
    links: list[TokenLink] = []


class OrderInfo(BaseModel):
    type: str
    status: str
    payment_timestamp: int = Field(..., alias="paymentTimestamp")


from setuptools import find_packages, setup


def read_file(file):
	with open(file, "r") as fh:
		return fh.read()

# py -m build
# py -m twine upload dist/*


setup(
	name="dexscreener",
	packages=find_packages(),
	version="1.2",
	license="MIT",

	description="Python wrapper for the 'dexscreener.com' API",
	long_description=read_file("README.md"),
	long_description_content_type="text/markdown",

	author="Joshua Nixon",
	author_email="nixonjoshua98@gmail.com",

	url="https://github.com/nixonjoshua98/dexscreener",

	download_url="https://github.com/nixonjoshua98/dexscreener/releases",

	keywords=[
		"dexscreener",
		"crypto",
		"cryptocurrency",
		"bitcoin"
	],

	install_requires=[
		"requests",
		"pydantic",
		"certifi",
		"aiohttp"
	],

	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"Topic :: Software Development :: Build Tools",
	],

	python_requires='>=3.9'
)
